package com.bocloud.cms.booter.controller;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.intf.FusionCloudService;
import com.bocloud.cms.intf.external.RemoteManageOneService;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Pager;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.Common;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.servlet.http.HttpServletRequest;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import java.util.*;

/**
 * FileSystem
 *
 * @author liuyuanyuan
 * @version 5.6.0
 * @date 2022/2/21 18:01
 */
@Tag(name = "FusionCloud监控")
@RestController
@RequestMapping("/v1/fusioncloud")
public class FusionCloudController {

    @Autowired
    private FusionCloudService fusionCloudService;

    @Autowired
    private RemoteManageOneService remoteManageOneService;

    /**
     * 华为云云主机列表查询
     *
     * @param pager   分页
     * @param context 当前用户上下文
     * @return 宿主机列表
     */
    @GetMapping("/hosts")
    @Operation(summary = "宿主机列表展示")
    public GeneralResult listOfHost(Pager pager, @Value(Common.REQ_CONTEXT) RequestContext context, HttpServletRequest request) {
        return fusionCloudService.listOfHost(pager, context);
    }

    /**
     * 资源池列表
     *
     * @param vendorId 平台id
     * @return 资源池列表
     */
    @GetMapping("/resource-pools")
    @Operation(summary = "资源池列表")
    public GeneralResult resourcePools(Long vendorId) {
        List<JSONObject> resourcePools = remoteManageOneService.listInstance(vendorId, "SYS_ResourcePool", null, null);
        List<JSONObject> azones = remoteManageOneService.listInstance(vendorId, "SYS_Azone", null, null);
        List<Map<String, Object>> resourcePoolList = new ArrayList<>();
        Optional.ofNullable(resourcePools).orElse(new ArrayList<>()).forEach(resourcePool -> {
            Map<String, Object> resourcePoolMap = new HashMap<>(16, 0.75F);
            resourcePoolMap.put("label", resourcePool.getString("name"));
            resourcePoolMap.put("value", resourcePool.getString("resId"));
            resourcePoolMap.put("tag", "resourcePool");
            List<Map<String, Object>> children = new ArrayList<>();
            resourcePoolMap.put("children", children);
            Optional.ofNullable(azones).orElse(new ArrayList<>()).forEach(azone -> {
                Map<String, Object> azoneMap = new HashMap<>(16, 0.75F);
                azoneMap.put("label", azone.getString("name"));
                azoneMap.put("value", azone.getString("resId"));
                azoneMap.put("tag", "azone");
                children.add(azoneMap);
            });
            resourcePoolList.add(resourcePoolMap);
        });
        return new GeneralResult<>(true, resourcePoolList, "success");
    }

    /**
     * 资源池详情
     *
     * @param id       资源池id
     * @param vendorId 平台id
     * @return 资源池实体类详情
     */
    @GetMapping("/resource-pools/{id}")
    @Operation(summary = "资源池详情")
    public GeneralResult detailResPool(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.getInstanceDetail(vendorId, "SYS_ResourcePool", id), "success");
    }

    /**
     * 可用去详情
     *
     * @param id       可用区id
     * @param vendorId 平台id
     * @return 可用去详情
     */
    @GetMapping("/azones/{id}")
    @Operation(summary = "可用区详情")
    public GeneralResult detailAzones(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.getInstanceDetail(vendorId, "SYS_Azone", id), "success");
    }

    /**
     * 查询资源池的云主机、物理机状态数目统计
     *
     * @param id       资源池id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/resource-pools/{id}/statistics")
    @Operation(summary = "查询资源池的云主机、物理机状态数目统计")
    public GeneralResult resourcePoolStatistics(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.resourcePoolStatistics(vendorId, id), "success");
    }

    /**
     * 查询资源池的云主机、物理机状态数目统计
     *
     * @param id       资源池id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/azones/{id}/statistics")
    @Operation(summary = "查询可用区的云主机、物理机状态数目统计")
    public GeneralResult azoneStatistics(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.azoneStatistics(vendorId, id), "success");
    }

    /**
     * 资源池的当前容量
     *
     * @param id       资源池id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/resource-pools/{id}/current-capacities")
    @Operation(summary = "资源池的当前容量")
    public GeneralResult resourcePoolCurrentCapacities(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.resourcePoolCurrentCapacities(vendorId, id), "success");
    }

    /**
     * 可用区的当前容量
     *
     * @param id       可用区id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/azones/{id}/current-capacities")
    @Operation(summary = "可用区的当前容量")
    public GeneralResult azoneCurrentCapacities(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.azoneCurrentCapacities(vendorId, id), "success");
    }

    /**
     * 资源池存储资源当前容量
     *
     * @param id       资源池id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/resource-pools/{id}/resource-types/storage-pool/capacities")
    @Operation(summary = "资源池存储资源当前容量")
    public GeneralResult resourcePoolStoragePoolCapacities(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.resourcePoolStoragePoolCapacities(vendorId, id), "success");
    }

    /**
     * 可用区存储资源当前容量
     *
     * @param id       可用去id
     * @param vendorId 平台id
     * @return 统计数据
     */
    @GetMapping("/capacity/azones/{id}/resource-types/storage-pool/capacities")
    @Operation(summary = "可用区存储资源当前容量")
    public GeneralResult azoneStoragePoolCapacities(@PathVariable String id, Long vendorId) {
        return new GeneralResult<>(true, remoteManageOneService.azoneStoragePoolCapacities(vendorId, id), "success");
    }
}
