package com.bocloud.cms.booter.listener;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.common.utils.PrometheusUtils;
import com.bocloud.cms.entity.Alarm;
import com.bocloud.cms.entity.Metric;
import com.bocloud.cms.entity.Rule;
import com.bocloud.cms.entity.RuleGroup;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.intf.*;
import com.bocloud.cms.model.PrometheusAlarm;
import com.bocloud.cms.model.PrometheusResult;
import com.bocloud.cms.repository.AlarmRepository;
import com.bocloud.cms.repository.MonitorResourceTypeRepository;
import com.bocloud.cms.repository.RuleGroupRepository;
import com.bocloud.cms.repository.RuleRepository;
import com.bocloud.cms.service.AlarmConfiguration;
import com.bocloud.cms.service.MonitoringAbleRegistry;
import com.bocloud.cms.service.core.AlarmExecutor;
import com.google.common.collect.Maps;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.lock.LockFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.MessageListener;
import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.stream.Collectors;
import java.util.concurrent.TimeUnit;

/**
 * 资源告警接收监听
 *
 * @author wangyu
 * @version 1.0
 * @since 2017年8月18日
 */
@Component
@Slf4j
public class AlarmListener implements MessageListener {

    ThreadFactory nameThreadFactory = new ThreadFactoryBuilder().setNameFormat("alarm-listener-pool-%d").build();

    private ExecutorService executor = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors() + 1, Runtime.getRuntime().availableProcessors() + 1, 0L, TimeUnit.MILLISECONDS,
            new LinkedBlockingQueue<>(), nameThreadFactory);

    private final RuleRepository ruleRepository;
    private final RuleGroupRepository ruleGroupRepository;
    private final PrometheusService prometheusService;
    private final AmqpTemplate amqpTemplate;
    private final MetricService metricService;
    private final AlarmService alarmService;
    private final AlarmSendPolicyService alarmSendPolicyService;
    private final AlarmNoticeService alarmNoticeService;
    private final AlarmRepository alarmRepository;
    private final LockFactory lockFactory;
    private final MonitorResourceTypeRepository monitorResourceTypeRepository;
    private final AlarmConfiguration alarmConfiguration;

    @Autowired
    public AlarmListener(RuleRepository ruleRepository, RuleGroupRepository ruleGroupRepository, MetricService metricService,
                         PrometheusService prometheusService, AmqpTemplate amqpTemplate, AlarmService alarmService,
                         AlarmSendPolicyService alarmSendPolicyService, AlarmNoticeService alarmNoticeService,
                         AlarmRepository alarmRepository, LockFactory lockFactory,
                         MonitorResourceTypeRepository monitorResourceTypeRepository,AlarmConfiguration alarmConfiguration) {
        this.ruleRepository = ruleRepository;
        this.ruleGroupRepository = ruleGroupRepository;
        this.prometheusService = prometheusService;
        this.amqpTemplate = amqpTemplate;
        this.metricService = metricService;
        this.alarmService = alarmService;
        this.alarmSendPolicyService = alarmSendPolicyService;
        this.alarmNoticeService = alarmNoticeService;
        this.alarmRepository = alarmRepository;
        this.lockFactory = lockFactory;
        this.monitorResourceTypeRepository = monitorResourceTypeRepository;
        this.alarmConfiguration = alarmConfiguration;
    }

    @RabbitListener(queues = "resource.alarm.queue", admin = "rabbitAdmin", autoStartup = "true")
    @Override
    public void onMessage(Message message) {
        try {
            String content = new String(message.getBody());
            log.info("receive alertmanager alert:" + content);
            PrometheusAlarm prometheusAlarm = JSONObject.parseObject(content, PrometheusAlarm.class);
            if (null == prometheusAlarm) {
                log.warn("get unexpected message, drop it!");
                return;
            }
            List<PrometheusAlarm.PrometheusAlert> nodeEventAlerts = prometheusAlarm.getAlerts().stream().filter(alert -> "nodeEvent".equalsIgnoreCase(String.valueOf(alert.getLabels().get("alertRuleType")))).collect(Collectors.toList());
            List<PrometheusAlarm.PrometheusAlert> performanceAlerts = prometheusAlarm.getAlerts().stream().filter(alert -> !"nodeEvent".equalsIgnoreCase(String.valueOf(alert.getLabels().get("alertRuleType")))).collect(Collectors.toList());
            // 数据库中间件节点改变事件告警
            if (!ListTool.isEmpty(nodeEventAlerts)) {
                handleNodeEvent(nodeEventAlerts);
            }
            // 资源性能告警
            if (!ListTool.isEmpty(performanceAlerts)) {
                handlePerformanceAlerts(performanceAlerts);
            }
        } catch (RuntimeException e) {
            log.error("parse alarm receive exception ", e);
        }
    }

    /**
     * 处理数据库中间件节点改变事件消息
     *
     * @param alerts 告警
     */
    private void handleNodeEvent(List<PrometheusAlarm.PrometheusAlert> alerts) {
        executor.submit(() -> {
            List<Map<String, Object>> list = alerts.stream().map(alert -> {
                Map<String, Object> map = MapTools.simpleMap("id", alert.getLabels().get("id"));
                map.put("category", alert.getLabels().get("category"));
                return map;
            }).collect(Collectors.toList());
            amqpTemplate.convertAndSend("software.node.event", list);
        });
    }

    /**
     * 处理资源性能告警
     *
     * @param alerts 告警
     */
    private void handlePerformanceAlerts(List<PrometheusAlarm.PrometheusAlert> alerts) {
        alerts.forEach(alert -> {
            Rule rule = ruleRepository.get(Rule.class, Long.valueOf(String.valueOf(alert.getLabels().get("ruleId"))));
            RuleGroup ruleGroup = ruleGroupRepository.get(RuleGroup.class, rule.getGroupId());
            //关闭状态时产生告警消息淹没
            if (!ruleGroup.getEnable() && "firing".equals(alert.getStatus())) {
                return;
            }
            Metric metric = metricService.query(rule.getMetricId());
            MonitoringAble monitoringAble = MonitoringAbleRegistry.getInstance(MonitorResourceType.valueOf(ruleGroup.getResourceType()));
            //从prometheus告警信息中解析资源对象
            Object object = monitoringAble.parseObjectFromAlert(alert, ruleGroup, rule, metric);
            //校验当前资源是否屏蔽告警
            if ("firing".equals(alert.getStatus())) {
                boolean alarmEnable = monitoringAble.alarmEnable(ruleGroup, object);
                if (!alarmEnable) {
                    return;
                }
            }
            Object value = alert.getAnnotations().get("value");
            Alarm alarm = new Alarm();
            alarm.setEventAlarm(false);
            alarm.setTenantId(ruleGroup.getTenantId());
            alarm.setName(String.valueOf(alert.getLabels().get("metricName")));
            alarm.setRuleId(rule.getId());
            alarm.setLevel(String.valueOf(alert.getLabels().get("level")));
            alarm.setType(ruleGroup.getResourceType());
            Object description = alert.getAnnotations().get("description");
            //告警触发时间为格林威治时间 慢8小时
            Date startsAt = alert.getStartsAt();
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(startsAt);
            calendar.add(Calendar.HOUR, 8);
            alarm.setGmtTrigger(calendar.getTime());
            // 设置告警个性化内容
            monitoringAble.wrapAlarm(alarm, object);
            Map<String, Object> props = Maps.newHashMap();
            props.put("metric", alert.getLabels().get("metricName"));
            props.put("expr", alert.getAnnotations().get("expr"));
            alarm.setProps(JSONObject.toJSONString(props));
            if (!metric.getIsEvent()) {
                alarm.setAlarmVal(value + metric.getUnit());
            }
            String status = alert.getStatus();
            //保存告警恢复的数值
            if ("resolved".equals(status)) {
                // update 告警恢复时截取当前时间时间戳为恢复时间
                long l = System.currentTimeMillis();
                //截时间戳 单位保留到s 未截取前单位为ms
                String endTime = Long.toString(l).substring(0, 10);
                String expr = String.valueOf(alert.getAnnotations().get("expr"));
                List<String> operators = Arrays.asList(">", "<", ">=", "<=", "==", "!=");
                for (String s : operators) {
                    expr = StringUtils.substringBeforeLast(expr, s);
                }
                // add 如果是全局告警 则在告警恢复时对其表达式进行拼接操作;
                if (ruleGroup.getIsGlobal()) {
                    Map<String, String> params = monitoringAble.getPrometheusExprParam(object, metric);
                    expr = PrometheusUtils.alarmRecoverPutParams(expr, params, ruleGroup.getResourceType());
                }
                List<PrometheusResult> list = prometheusService.query(expr, endTime);
                log.info("Global alarm recovery exporter:" + expr);
                log.info("alertEndTime:" + alert.getEndsAt().getTime());
                log.info("endTime:" + endTime);
                if (!list.isEmpty()) {
                    if (!metric.getIsEvent()) {
                        // 如果为指标告警 则正常拼接告警恢复值
                        Object recoveryValue = list.get(0).getValue()[1];
                        alarm.setRecoveryVal(recoveryValue + metric.getUnit());
                    }
                } else {
                    alarm.setRecoveryVal("未知");
                }
            }
            // 如果为事件告警则不拼接当前值类型等信息
            if (metric.getIsEvent()) {
                alarm.setRemark(description.toString());
            } else {
                alarm.setRemark(description + "，当前值" + value + metric.getUnit());
            }
            AlarmExecutor alarmExecutor = new AlarmExecutor(alarm, object, metric, status, alarmService, ruleRepository, ruleGroupRepository,
                    alarmSendPolicyService, alarmNoticeService, alarmRepository, lockFactory, monitorResourceTypeRepository,
                    alarmConfiguration);
            executor.submit(alarmExecutor);
        });
    }
}
