package com.bocloud.cms.repository;

import com.bocloud.cms.entity.ResourceAlarmStatus;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.MapTools;
import com.megatron.database.core.intf.impl.BasicGenericDao;
import com.megatron.framework.core.CurrentService;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * 开启关闭告警状态
 *
 * @author zhangyf
 * @version 5.6.0
 * @date 2022/2/21 18:01
 */
@Repository
public class ResourceAlarmStatusRepository extends BasicGenericDao<ResourceAlarmStatus, Long> {

    public ResourceAlarmStatusRepository(JdbcTemplate jdbcTemplate, NamedParameterJdbcTemplate npJdbcTemplate, CurrentService service) {
        super(jdbcTemplate, npJdbcTemplate, service);
    }

    /**
     * 资源告警状态集合
     *
     * @param resourceType 资源类型
     * @param targets      监控资源ID集合
     * @param context      当前用户上下文
     * @return 资源告警状态集合
     */
    public List<ResourceAlarmStatus> list(String resourceType, List<String> targets, RequestContext context) {
        if (targets == null || targets.isEmpty()) {
            return new ArrayList<>();
        }
        String builder = "select * from monitor_resource_alarm_status where resource_type = :resourceType and tenant_id = :tenantId " +
                "  and resource_id in (:resourceId) ";
        Map<String, Object> paramMap = MapTools.simpleMap("resourceType", resourceType);
        paramMap.put("tenantId", context.getTenant());
        if (targets.size() > 0) {
            paramMap.put("resourceId", targets);
        }
        return this.list(ResourceAlarmStatus.class, builder, paramMap);
    }

    /**
     * 根据id和属性删除对象
     *
     * @param resourceId   监控资源ID
     * @param resourceType 资源类型
     * @param context      当前用户上下文
     * @return 是否成功
     */
    public boolean delete(String resourceId, String resourceType, RequestContext context) {
        String sql = "delete from monitor_resource_alarm_status where resource_type = :resourceType  and resource_id = :resourceId and tenant_id = :tenantId ";
        Map<String, Object> paramMap = MapTools.simpleMap("resourceType", resourceType);
        if (resourceId != null) {
            paramMap.put("resourceId", resourceId);
        }
        paramMap.put("tenantId", context.getTenant());
        return this.execute(sql, paramMap) > 0;
    }

    /**
     * 是否开启告警(默认开启)
     *
     * @param resourceType 资源类型
     * @param target       监控资源ID
     * @param tenantId     租户ID
     * @return 是否开启
     */
    public boolean alarmEnable(String resourceType, String target, Long tenantId) {
        String sql = "select * from monitor_resource_alarm_status where resource_type = :resourceType " +
                " and resource_id = :resourceId and tenant_id = :tenantId ";
        Map<String, Object> paramMap = MapTools.simpleMap("resourceType", resourceType);
        paramMap.put("resourceId", target);
        paramMap.put("tenantId", tenantId);
        List<ResourceAlarmStatus> list = this.list(ResourceAlarmStatus.class, sql, paramMap);
        return !Optional.ofNullable(list).orElse(new ArrayList<>()).stream().findAny().isPresent();
    }

}
