package com.bocloud.cms.intf;

import com.bocloud.cms.entity.Alarm;
import com.bocloud.cms.entity.Metric;
import com.bocloud.cms.entity.Rule;
import com.bocloud.cms.entity.RuleGroup;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.model.PrometheusAlarm;

import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Map;

/**
 * 告警资源对象接口
 *
 * @author zhangyf
 * @since 2019-12-23
 */
public interface MonitoringAble<T> {

    /**
     * 获取默认的唯一标识，如果实现类T中不是使用id作为唯一标识，则需要复写此方法
     *
     * @param target       监控资源目标对象
     * @param resourceType 资源类型
     * @return 监控目标对象ID
     * @throws NoSuchMethodException     没有getId方法
     * @throws InvocationTargetException 反射异常
     * @throws IllegalAccessException    参数错误
     */
    default String getTargetObjectId(T target, MonitorResourceType resourceType) throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        return String.valueOf(target.getClass().getMethod("getId").invoke(target));
    }

    /**
     * 获取告警目标对象
     *
     * @param targetIds    目标资源id
     * @param resourceType 资源类型
     * @return 告警目标对象集合
     */
    List<T> listTargetObjects(List<String> targetIds, MonitorResourceType resourceType);

    /**
     * 获取表达式参数<br />
     * 用于生成告警规则时匹配资源
     *
     * @param targetObject 告警目标对象
     * @param metric       监控指标
     * @return Prometheus表达式参数
     */
    Map<String, String> getPrometheusExprParam(T targetObject, Metric metric);

    /**
     * 从prometheus告警信息中解析资源对象
     *
     * @param alert     PrometheusAlert
     * @param ruleGroup 告警策略
     * @param rule      告警规则
     * @param metric    监控指标
     * @return 监控目标资源对象
     */
    T parseObjectFromAlert(PrometheusAlarm.PrometheusAlert alert, RuleGroup ruleGroup, Rule rule, Metric metric);

    /**
     * 资源被删除
     *
     * @param object 被删除资源对象
     * @throws Exception 处理被删除资源对象回调时异常
     */
    default void destroy(T object) throws Exception {
        destroy(object, null);
    }


    /**
     * 资源被删除
     *
     * @param object  被删除资源对象
     * @param catalog 资源类型
     * @throws Exception 处理被删除资源对象回调时异常
     */
    void destroy(T object, String catalog) throws Exception;

    /**
     * 接收到告警后设置对应资源个性化内容
     *
     * @param alarm  告警
     * @param object parseObjectFromAlert 方法返回的对象
     */
    void wrapAlarm(Alarm alarm, T object);

    /**
     * 资源告警接收状态
     *
     * @param ruleGroup 告警策略
     * @param object    监控资源对象
     * @return 是否接收资源告警
     */
    boolean alarmEnable(RuleGroup ruleGroup, T object);

}
