package com.bocloud.cms.service;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.common.config.StorageConfig;
import com.bocloud.cms.entity.*;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.enums.ResourceCategory;
import com.bocloud.cms.intf.*;
import com.bocloud.cms.model.CcsModel;
import com.bocloud.cms.model.PrometheusAlarm;
import com.bocloud.cms.repository.ResourceAlarmStatusRepository;
import com.bocloud.cms.service.internal.mms.MmsCcsService;
import com.bocloud.pms.entity.Ccs;
import com.bocloud.pms.entity.CcsNode;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Pager;
import com.megatron.common.model.Param;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.MapTools;
import jakarta.servlet.http.HttpServletRequest;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import com.megatron.common.model.*;


import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * service实现层
 * @author zhangyf
 * @version 1.0
 * @date 2022/2/21 18:01
 * @version 5.6.0
 */
@Service
@Slf4j
public class CcsServiceImpl implements CcsService, InitializingBean {


    @Autowired
    private MmsCcsService mmsCcsService;
    @Autowired
    private PrometheusService prometheusService;
    @Autowired
    private PrometheusRuleService prometheusRuleService;
    @Autowired
    private StorageConfig storageConfig;
    @Autowired
    private AlarmService alarmService;
    @Autowired
    private ResourceAlarmStatusRepository alarmStatusRepository;
    @Autowired
    private ResourceAlarmStatusService resourceAlarmStatusService;

    @Override
    public void init(Ccs ccs) throws Exception {
        switch (ResourceCategory.valueOf(ccs.getType())) {
            case Zookeeper:
                generateZookeeperConf(ccs);
                break;
            case Etcd:
                generateEtcdConf(ccs);
                break;
            default:
                log.warn("unsupported resource category " + ccs.getType());
                break;
        }
    }

    @Override
    public List<Ccs> listTargetObjects(List<String> targetIds, MonitorResourceType resourceType) {
        List<Integer> targetList = new ArrayList<>();
        Optional.ofNullable(targetIds).orElse(new ArrayList<>()).forEach(targetId -> targetList.add(Integer.valueOf(targetId)));
        Param param = new Param(MapTools.simpleMap("id", JSONObject.toJSONString(targetList).substring(1, JSONObject.toJSONString(targetList).length() - 1)), Sign.IN);

        List<Param> params = Collections.singletonList(param);
        GeneralResult result = mmsCcsService.list(new Pager(1, Integer.MAX_VALUE, params, null));
        Assert.isTrue(result.isSuccess(), result.getMessage());
        JSONObject data = (JSONObject) result.getData();
        return data.getJSONArray(Common.ROWS).toJavaList(Ccs.class);
    }

    @Override
    public Map<String, String> getPrometheusExprParam(Ccs targetObject, Metric metric) {
        Map<String, String> params = new HashMap<>(16, 0.75F);
        params.put("ccs_id", String.valueOf(targetObject.getId()));
        return params;
    }

    @Override
    public Ccs parseObjectFromAlert(PrometheusAlarm.PrometheusAlert alert, RuleGroup ruleGroup, Rule rule, Metric metric) {
        String ccsId = String.valueOf(alert.getLabels().get("ccs_id"));
        return mmsCcsService.detail(Long.parseLong(ccsId));
    }

    @Override
    public void destroy(Ccs ccs, String catalog) throws Exception {
        //删除prometheus targets文件
        File file = new File(new File(storageConfig.getPrometheusPath(), ccs.getType().toLowerCase()), "instance_" + ccs.getId() + ".json");
        if (file.exists() && !file.delete()) {
            log.error("file delete failed, path:{} " + file.getPath());
        }
        // 删除实例节点状态监听rule文件
        File ruleFile = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_ccs_" + ccs.getId() + ".yml");
        if (ruleFile.exists() && !ruleFile.delete()) {
            log.error("file delete failed, path:{} " + ruleFile.getPath());
        }
        prometheusService.reload();
        //TODO
    }

    @Override
    public void wrapAlarm(Alarm alarm, Ccs ccs) {
        alarm.setTarget(ccs.getId());
        if (ccs.getName()!=null & ccs.getManageIp()!=null){
            alarm.setTargetName(ccs.getName() + "(" + ccs.getManageIp() + ")");
        }else{
            String targetName = ccs.getManageIp() == null ? ccs.getName() : ccs.getManageIp();
            alarm.setTargetName(targetName);
        }
        alarm.setOwnerId(ccs.getCreatorId());
    }

    @Override
    public boolean alarmEnable(RuleGroup ruleGroup, Ccs object) {
        return resourceAlarmStatusService.alarmEnable(ruleGroup.getResourceType(), String.valueOf(object.getId()), ruleGroup.getTenantId());
    }

    @Override
    public GeneralResult list(Pager pager, String catalog, RequestContext context, HttpServletRequest request) {

        GeneralResult result = mmsCcsService.list(pager, request.getHeader(Common.TOKEN));
        if (result.isFailed()) {
            return result;
        }
        JSONObject data = (JSONObject) result.getData();
        List<CcsModel> list = data.getJSONArray(Common.ROWS).toJavaList(CcsModel.class);
        if (list.isEmpty()) {
            data.put(Common.ROWS, list);
            return result;
        }
        List<String> targets = list.stream().map(m -> String.valueOf(m.getId())).collect(Collectors.toList());
        List<ResourceAlarmStatus> alarmStatuses = this.alarmStatusRepository.list(catalog, targets, context);
        for (CcsModel model : list) {
            //告警启用状态，默认开启
            boolean alarmEnable = !Optional.ofNullable(alarmStatuses).orElse(new ArrayList<>()).stream()
                    .filter(a -> a.getResourceId().equals(model.getId().toString()))
                    .findAny().isPresent();
            model.setAlarmEnable(alarmEnable);
        }
        data.put(Common.ROWS, list);
        return result;
    }


    /**
     * 生成Zookeeper配置
     *
     * @param ccs
     * @throws IOException
     */
    private void generateZookeeperConf(Ccs ccs) throws IOException {
        // 创建prometheus targets文件
        List<JSONObject> list = ccs.getNodes().stream().map(node -> {
            Map<String, Object> labels = new HashMap<>(16, 0.75F);
            labels.put("ccs_id", String.valueOf(ccs.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("hostname", node.getAddress() + ":" + node.getPort());
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9273"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.Zookeeper.name().toLowerCase()), "instance_" + ccs.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "zookeeper node change event");
        JSONArray rules = new JSONArray();
        StringBuilder expr = new StringBuilder("changes((count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "'}) + 1 or absent(count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "'})))[3m:]) > 0 ");
        for (CcsNode node : ccs.getNodes()) {
            expr.append(" or changes((count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "',node_id='" + node.getId() + "',state='leader'}) + 1 ");
            expr.append(" or absent(count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "',node_id='" + node.getId() + "',state='leader'})))[3m:]) > 0 ");
            expr.append(" or changes((count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "',node_id='" + node.getId() + "',state='follow'}) + 1 ");
            expr.append(" or absent(count(zookeeper_num_alive_connections{ccs_id='" + ccs.getId() + "',node_id='" + node.getId() + "',state='follow'})))[3m:]) > 0");
        }
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.Zookeeper.name());
        labels.put("id", ccs.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "zookeeper node change event");
        rule.put("expr", expr.toString());
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "zookeeper instance(" + ccs.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_ccs_" + ccs.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    /**
     * 生成Etcd配置
     *
     * @param ccs
     * @throws IOException
     */
    private void generateEtcdConf(Ccs ccs) throws IOException {
        // 创建prometheus targets文件
        List<JSONObject> list = ccs.getNodes().stream().map(node -> {
            Map<String, Object> labels = new HashMap<>(16, 0.75F);
            labels.put("ccs_id", String.valueOf(ccs.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("hostname", node.getAddress() + ":" + node.getPort());
            JSONObject config = new JSONObject();
            //使用自身监控，端口号2379
            config.put("targets", Arrays.asList(node.getManageIp() + ":2379"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.Etcd.name().toLowerCase()), "instance_" + ccs.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "etcd node change event");
        JSONArray rules = new JSONArray();
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.Etcd.name());
        labels.put("id", ccs.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "etcd node change event");
        rule.put("expr", "changes((count(etcd_server_health_success{ccs_id='" + ccs.getId() + "'}) + 1 or absent(count(etcd_server_health_success{ccs_id='" + ccs.getId() + "'})))[3m:]) > 0 ");
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "etcd instance(" + ccs.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_ccs_" + ccs.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_CCS_ZOOKEEPER, this);
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_CCS_ETCD, this);
    }

}
