package com.bocloud.cms.service;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.common.config.StorageConfig;
import com.bocloud.cms.entity.*;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.enums.ResourceCategory;
import com.bocloud.cms.intf.CmsService;
import com.bocloud.cms.intf.PrometheusRuleService;
import com.bocloud.cms.intf.PrometheusService;
import com.bocloud.cms.intf.ResourceAlarmStatusService;
import com.bocloud.cms.model.CmsModel;
import com.bocloud.cms.model.PrometheusAlarm;
import com.bocloud.cms.model.PrometheusResult;
import com.bocloud.cms.repository.ResourceAlarmStatusRepository;
import com.bocloud.cms.service.internal.mms.MmsCmsService;
import com.bocloud.pms.entity.Cms;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.megatron.common.model.*;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.MapTools;
import jakarta.servlet.http.HttpServletRequest;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import com.megatron.common.model.*;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * cmsService实现类
 * @author zhangyf
 * @version 1.0
 * @date 2022/2/21 18:01
 * @version 5.6.0
 */
@Service
@Slf4j
public class CmsServiceImpl implements CmsService, InitializingBean {


    @Autowired
    private MmsCmsService mmsCmsService;
    @Autowired
    private PrometheusService prometheusService;
    @Autowired
    private StorageConfig storageConfig;
    @Autowired
    private PrometheusRuleService prometheusRuleService;
    @Autowired
    private ResourceAlarmStatusRepository alarmStatusRepository;
    @Autowired
    private ResourceAlarmStatusService resourceAlarmStatusService;

    @Override
    public void init(Cms cms) throws Exception {
        switch (ResourceCategory.valueOf(cms.getType())) {
            case Kafka:
                generateKafkaConf(cms);
                break;
            case RabbitMQ:
                generateRabbitmqConf(cms);
                break;
            default:
                log.warn("unsupported resource category " + cms.getType());
                break;
        }
    }

    @Override
    public List<Cms> listTargetObjects(List<String> targetIds, MonitorResourceType resourceType) {
        List<Integer> targetList = new ArrayList<>();
        Optional.ofNullable(targetIds).orElse(new ArrayList<>()).forEach(targetId -> targetList.add(Integer.valueOf(targetId)));
        Param param = new Param(MapTools.simpleMap("id", JSONObject.toJSONString(targetList).substring(1, JSONObject.toJSONString(targetList).length() - 1)), Sign.IN);
        List<Param> params = Collections.singletonList(param);
        GeneralResult result = mmsCmsService.list(new Pager(1, Integer.MAX_VALUE, params, null));
        Assert.isTrue(result.isSuccess(), result.getMessage());
        JSONObject data = (JSONObject) result.getData();
        return data.getJSONArray(Common.ROWS).toJavaList(Cms.class);
    }

    @Override
    public Map<String, String> getPrometheusExprParam(Cms targetObject, Metric metric) {
        Map<String, String> params = new HashMap<>(16, 0.75F);
        params.put("cms_id", String.valueOf(targetObject.getId()));
        return params;
    }

    @Override
    public Cms parseObjectFromAlert(PrometheusAlarm.PrometheusAlert alert, RuleGroup ruleGroup, Rule rule, Metric metric) {
        String cmsId = String.valueOf(alert.getLabels().get("cms_id"));
        return mmsCmsService.detail(Long.parseLong(cmsId));
    }

    @Override
    public void destroy(Cms cms, String catalog) throws Exception {
        // 资源回收后删除配置文件
        File file = new File(new File(storageConfig.getPrometheusPath(), cms.getType().toLowerCase()), "instance_" + cms.getId() + ".json");
        if (file.exists() && !file.delete()) {
            log.error("file delete failed, path:{} " + file.getPath());
        }
        // 删除实例节点状态监听rule文件
        File ruleFile = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_cms_" + cms.getId() + ".yml");
        if (ruleFile.exists() && !ruleFile.delete()) {
            log.error("file delete failed, path:{} " + ruleFile.getPath());
        }
        prometheusService.reload();

        //TODO
    }

    @Override
    public void wrapAlarm(Alarm alarm, Cms cms) {
        alarm.setTarget(cms.getId());
        if (cms.getName()!=null & cms.getManageIp()!=null){
            alarm.setTargetName(cms.getName() + "(" + cms.getManageIp() + ")");
        }else{
            String targetName = cms.getManageIp() == null ? cms.getName() : cms.getManageIp();
            alarm.setTargetName(targetName);
        }
        alarm.setOwnerId(cms.getCreatorId());
    }

    @Override
    public boolean alarmEnable(RuleGroup ruleGroup, Cms object) {
        return resourceAlarmStatusService.alarmEnable(ruleGroup.getResourceType(), String.valueOf(object.getId()), ruleGroup.getTenantId());
    }

    @Override
    public GeneralResult list(Pager pager, String catalog, RequestContext context, HttpServletRequest request) {

        GeneralResult result = mmsCmsService.list(pager, request.getHeader(Common.TOKEN));
        if (result.isFailed()) {
            return result;
        }
        JSONObject data = (JSONObject) result.getData();
        List<CmsModel> list = data.getJSONArray(Common.ROWS).toJavaList(CmsModel.class);
        if (list.isEmpty()) {
            data.put(Common.ROWS, list);
            return result;
        }
        List<String> targets = list.stream().map(m -> String.valueOf(m.getId())).collect(Collectors.toList());
        List<ResourceAlarmStatus> alarmStatuses = this.alarmStatusRepository.list(catalog, targets, context);
        for (CmsModel model : list) {
            //告警启用状态，默认开启
            boolean alarmEnable = !Optional.ofNullable(alarmStatuses).orElse(new ArrayList<>()).stream()
                    .filter(a -> a.getResourceId().equals(model.getId().toString()))
                    .findAny().isPresent();
            model.setAlarmEnable(alarmEnable);
        }
        data.put(Common.ROWS, list);
        return result;
    }

    @Override
    public List<SimpleBean> topicsOfKafka(Long id) {
        List<PrometheusResult> list = prometheusService.query("kafka_server_brokertopicmetrics_messagesin_total{topic!='',cms_id='" + id + "'}");
        Set<String> set = list.stream().map(prometheusResult -> (String) prometheusResult.getMetric().get("topic")).collect(Collectors.toSet());
        return Optional.ofNullable(set).orElse(new HashSet<>()).stream().map(value -> new SimpleBean(value, value)).collect(Collectors.toList());
    }

    @Override
    public List<SimpleBean> consumerGroupsOfKafka(Long id) {
        List<PrometheusResult> list = prometheusService.query("sum(kafka_consumergroup_lag{cms_id='" + id + "'}) by (consumergroup)");
        Set<String> set = list.stream().map(prometheusResult -> (String) prometheusResult.getMetric().get("consumergroup")).collect(Collectors.toSet());
        return Optional.ofNullable(set).orElse(new HashSet<>()).stream().map(value -> new SimpleBean(value, value)).collect(Collectors.toList());
    }

    /**
     * 创建Kafka配置文件
     *
     * @param cms
     * @throws IOException
     */
    private void generateKafkaConf(Cms cms) throws IOException {
        // 创建prometheus targets文件
        // telegraf，jmx_exporter、kafka_exporter
        List<JSONObject> list = cms.getNodes().stream().map(node -> {
            Map<String, String> labels = new HashMap<>(16, 0.75F);
            labels.put("cms_id", String.valueOf(cms.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("server", node.getAddress());
            labels.put("port", String.valueOf(node.getPort()));
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9990", node.getManageIp() + ":9308"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.Kafka.name().toLowerCase()), "instance_" + cms.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "kafka node change event");
        JSONArray rules = new JSONArray();
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.Kafka.name());
        labels.put("id", cms.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "kafka node change event");
        rule.put("expr", "changes((count(kafka_server_brokertopicmetrics_messagesin_total{cms_id='" + cms.getId() + "',topic=''}) + 1 or absent(count(kafka_server_brokertopicmetrics_messagesin_total{cms_id='" + cms.getId() + "',topic=''})))[3m:]) > 0");
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "kafka instance(" + cms.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_cms_" + cms.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    /**
     * 生成telegraf RabbitMQ配置
     *
     * @param cms
     * @throws IOException
     */
    private void generateRabbitmqConf(Cms cms) throws Exception {
        // 创建prometheus targets文件
        List<JSONObject> list = cms.getNodes().stream().map(node -> {
            Map<String, String> labels = new HashMap<>(16, 0.75F);
            labels.put("cms_id", String.valueOf(cms.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("server", node.getAddress() + ":" + node.getPort());
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9273"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.RabbitMQ.name().toLowerCase()), "instance_" + cms.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "rabbitmq node change event");
        JSONArray rules = new JSONArray();
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.RabbitMQ.name());
        labels.put("id", cms.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "rabbitmq node change event");
        rule.put("expr", "changes((count(rabbitmq_overview_connections{cms_id='" + cms.getId() + "'}) + 1 or absent(count(rabbitmq_overview_connections{cms_id='" + cms.getId() + "'})))[3m:]) > 0");
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "rabbitmq instance(" + cms.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_cms_" + cms.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_CMS_KAFKA, this);
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_CMS_RABBITMQ, this);
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_CMS_ROCKETMQ, this);
    }


}
