package com.bocloud.cms.service;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.bocloud.cms.common.config.StorageConfig;
import com.bocloud.cms.common.utils.Constant;
import com.bocloud.cms.entity.MonitorAllocationRatio;
import com.bocloud.cms.entity.MonitorHostIpmi;
import com.bocloud.cms.intf.HostService;
import com.bocloud.cms.intf.PrometheusService;
import com.bocloud.cms.model.ServerModel;
import com.bocloud.cms.repository.AllocationRatioRepository;
import com.bocloud.cms.repository.MonitorHostIpmiRepository;
import com.bocloud.cms.service.adapter.HostAdapter;
import com.bocloud.cms.service.internal.CmpCloudVendorService;
import com.bocloud.cms.service.internal.CmpServerService;
import com.bocloud.ims.entity.CloudVendor;
import com.bocloud.ims.entity.Server;
import com.bocloud.ims.entity.ServerConfig;
import com.bocloud.ims.enums.CloudProvider;
import com.google.common.collect.Maps;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.megatron.common.encrypt.AESEncryptor;
import com.megatron.common.http.HttpClient;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Pager;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.Common;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import org.yaml.snakeyaml.Yaml;

import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * 主机Service实现类
 *
 * @author zhangyf
 * @version 5.6.0
 * @since 2022/2/21 18:01
 */
@RequiredArgsConstructor
@Service
@Slf4j
public class HostServiceImpl implements HostService {

    @Value("${ipmi.exporter.url}")
    private String ipmiExporterUrl;

    ThreadFactory generateRuleFactory = new ThreadFactoryBuilder().setNameFormat("host-service-pool-%d").build();
    private ExecutorService generateThreadPool = new ThreadPoolExecutor(Runtime.getRuntime().availableProcessors() + 1,
            Runtime.getRuntime().availableProcessors() + 1, 0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<>(), generateRuleFactory);

    private final CmpServerService cmpServerService;
    private final CmpCloudVendorService cmpCloudVendorService;
    private final PrometheusService prometheusService;
    private final AllocationRatioRepository allocationRatioRepository;
    private final MonitorHostIpmiRepository monitorHostIpmiRepository;
    private final StorageConfig storageConfig;
    private final LockFactory lockFactory;
    private final ApplicationContext applicationContext;

    @Override
    public GeneralResult<?> list(Pager pager, RequestContext context) {
        GeneralResult<?> result = cmpServerService.list(pager);
        if (result.isFailed()) {
            return result;
        }
        JSONObject data = (JSONObject) result.getData();
        List<ServerModel> list = data.getJSONArray(Common.ROWS).toJavaList(ServerModel.class);
        if (list.isEmpty()) {
            return result;
        }
        for (ServerModel server : list) {
            server.setCpuTotal(server.getConfig().getCpuHz() != null ? String.valueOf(server.getConfig().getCpuHz()) : null);
            server.setMemTotal(server.getConfig().getMemoryCapacity() != null ? String.valueOf(server.getConfig().getMemoryCapacity()) : null);
        }
        Map<String, HostAdapter> adapters = applicationContext.getBeansOfType(HostAdapter.class);
        adapters.forEach((s, adapter) -> adapter.list(list, context));
        data.put(Common.ROWS, list);
        return result;
    }


    @Override
    public GeneralResult hostOverview(Long id, String type) {
        try {
            GeneralResult<Server> server = cmpServerService.detail(id);
            Assert.isTrue(server.isSuccess(), " 未查询到该主机相关信息");
            HashMap<String, Object> result = new HashMap<>(16, 0.75F);
            if (type.equals(CloudProvider.VMWARE.name())) {
                result.put("name", server.getData().getName());
                result.put("powerState", server.getData().getPowerState());
                result.put("ip", server.getData().getManageIp());
                result.put("vendor", server.getData().getConfig().getVendor());
                result.put("memoryCapacity", server.getData().getConfig().getMemoryCapacity());
                result.put("diskCapacity", server.getData().getConfig().getDiskCapacity());
                result.put("cpuCapacity", server.getData().getConfig().getCpuHz());
                result.put("vmNum", server.getData().getVmNum());
            } else if (type.equals(CloudProvider.OPENSTACK.name())) {
                GeneralResult<CloudVendor> cloudVendor = cmpCloudVendorService.detail(server.getData().getVendorId());
                String hostId = server.getData().getUuid();
                Long vendorId = cloudVendor.getData().getId();
                MonitorAllocationRatio ratio = allocationRatioRepository.query(cloudVendor.getData().getId());
                Double cpuCapacity = prometheusService
                        .simpleQuery("bocloud_openstack_hypervisor_vcpus_total{vendor_id='" + vendorId + "',id = '" + hostId + "'}");
                Double cpuUsed = prometheusService
                        .simpleQuery("bocloud_openstack_hypervisor_vcpus_used{vendor_id='" + vendorId + "',id = '" + hostId + "'}");
                Double memoryCapacity = prometheusService.simpleQuery("round(bocloud_openstack_hypervisor_memory_mbs_total{vendor_id='" + vendorId + "',id = '" + hostId + "'}/1024,0.01)");
                Double memoryUsed = prometheusService.simpleQuery("round(bocloud_openstack_hypervisor_memory_mbs_used{vendor_id='" + vendorId + "',id = '" + hostId + "'}/1024,0.01)");
                Double diskCapacity = prometheusService.simpleQuery("bocloud_openstack_hypervisor_disk_gbs_total{vendor_id='" + vendorId + "',id = '" + hostId + "'}");
                Double diskUsed = prometheusService.simpleQuery("bocloud_openstack_hypervisor_disk_gbs_used{vendor_id='" + vendorId + "',id = '" + hostId + "'}");
                Double cpuCore = prometheusService.simpleQuery("sum(count(node_cpu_seconds_total{host_id='" + server.getData().getId() + "', mode='system'}) by (cpu))");
                Double vmNum = prometheusService.simpleQuery("bocloud_openstack_hypervisor_running_vms{vendor_id='" + vendorId + "',id = '" + hostId + "'}");
                result.put("name", server.getData().getName());
                result.put("ip", server.getData().getManageIp());
                result.put("vmNum", vmNum);
                result.put("cpuUsed", cpuUsed);
                Long cpuAllocationRate = ratio != null ? ratio.getVcpuRatio() : 1L;
                result.put("cpuAllocationRate", cpuAllocationRate);
                result.put("cpuCapacity", cpuCapacity == null ? 0 : cpuCapacity * cpuAllocationRate);
                result.put("memUsed", memoryUsed);
                Long memAllocationRate = ratio != null ? ratio.getRamRatio() : 1L;
                result.put("memAllocationRate", memAllocationRate);
                result.put("memoryCapacity", memoryCapacity == null ? 0 : memoryCapacity * memAllocationRate);
                result.put("diskUsed", diskUsed);
                Long diskAllocationRate = ratio != null ? ratio.getDiskRatio() : 1L;
                result.put("diskAllocationRate", diskAllocationRate);
                result.put("diskCapacity", diskCapacity == null ? 0 : diskCapacity * diskAllocationRate);
                result.put("cpuCore", cpuCore);
            }
            return new GeneralResult<>(true, result, "查询成功");
        } catch (Exception e) {
            log.error("Query host overview information fail:", e);
            return new GeneralResult<>(false, "查询失败" + e.getMessage());
        }
    }

    @Override
    public GeneralResult hostInfo(Long id, String type) {
        GeneralResult<Server> server = cmpServerService.detail(id);
        Assert.isTrue(server.isSuccess(), " 未查询到该主机相关信息");
        HashMap<String, Object> result = new HashMap<>(16, 0.75F);
        DecimalFormat format = new DecimalFormat("0.##");
        ServerConfig serverConfig = Optional.ofNullable(server.getData()).map(d -> d.getConfig()).orElse(new ServerConfig());
        String name = Optional.ofNullable(server.getData()).map(d -> d.getName()).orElse(null);
        if ("cpu".equals(type)) {
            Double cpuUsed = prometheusService
                    .simpleQuery("vsphere_host_cpu_usagemhz_average{cpu='instance-total',source='" + name + "'}");
            // 查询主机cpu平均使用情况
            Double cpuUsage = prometheusService
                    .simpleQuery("vsphere_host_cpu_usage_average{cpu='instance-total',source='" + name + "'}");
            result.put("cpuUsed", cpuUsed != null ? cpuUsed : "");
            result.put("cpuUsage", cpuUsage != null ? cpuUsage : "");
            if (null != serverConfig) {
                result.put("cpuModel", serverConfig.getCpuMode());
                result.put("cpuAmount", serverConfig.getCpuAmount());
                result.put("cpuCore", serverConfig.getCpuCore());
                result.put("cpuTotal", serverConfig.getCpuHz());
                result.put("cpuFree", cpuUsed != null ? serverConfig.getCpuHz() - cpuUsed : null);
            }
        } else if ("mem".equals(type)) {
            Double memUsed = prometheusService
                    .simpleQuery("round(vsphere_host_mem_consumed_average{source='" + name + "'}/ 1024 / 1024,0.01)");
            result.put("memUsed", memUsed != null ? memUsed : "");
            result.put("memTotalCapacity", serverConfig != null ? serverConfig.getMemoryCapacity() : "");
            if (null != serverConfig && null != memUsed) {
                result.put("memFree", Double.parseDouble(format.format(serverConfig.getMemoryCapacity() - memUsed)));
            }
        } else if ("disk".equals(type)) {
            if (serverConfig != null) {
                result.put("diskCapacity", serverConfig.getDiskCapacity());
                result.put("diskUsed", serverConfig.getDiskUsed());
                Double diskFree = Double.parseDouble(format.format(serverConfig.getDiskCapacity() - serverConfig.getDiskUsed()));
                result.put("diskFree", diskFree);
            }
        }
        return new GeneralResult<>(true, result, "查询成功");
    }

    @Override
    public GeneralResult binding(MonitorHostIpmi monitorHostIpmi, RequestContext context) {
        MonitorHostIpmi hostIpmi = monitorHostIpmiRepository.query(monitorHostIpmi.getHostId());
        if (null == hostIpmi) {
            monitorHostIpmi.setGmtCreate(new Date());
            monitorHostIpmi.setDeleted(false);
            monitorHostIpmi.setCreatorId(context.getTarget());
            monitorHostIpmiRepository.save(monitorHostIpmi);
        } else {
            String path = MonitorHostIpmi.class.getSimpleName() + "_" + hostIpmi.getId();
            try (AutoCloseLock acLock = lockFactory.getACLock(path)) {
                Assert.isTrue(acLock.acquire(10, TimeUnit.SECONDS), "请求超时");
                monitorHostIpmi.setDeleted(false);
                monitorHostIpmi.setId(hostIpmi.getId());
                monitorHostIpmi.setGmtModify(new Date());
                monitorHostIpmi.setMenderId(context.getTarget());
                monitorHostIpmiRepository.update(monitorHostIpmi);
            } catch (Exception e1) {
                log.error("Configure the host for ipmi is error:", e1);
                return new GeneralResult<>(false, "配置失败");
            }
        }
        List<MonitorHostIpmi> monitorHostIpmis = monitorHostIpmiRepository.list();
        changeFile(monitorHostIpmis);
        reload();
        return new GeneralResult<>(true, "配置成功");
    }

    /**
     * 变更配置文件
     *
     * @param monitorHostIpmis
     */
    private void changeFile(List<MonitorHostIpmi> monitorHostIpmis) {
        //重新生成config.yml文件
        Map<String, Object> root = new LinkedHashMap<>();
        Map<Long, Object> data = new LinkedHashMap<>();
        JSONArray targets = new JSONArray();
        for (MonitorHostIpmi monitorHostIpmi : monitorHostIpmis) {
            GeneralResult<Server> serverGeneralResult = cmpServerService.detail(monitorHostIpmi.getHostId());
            if (serverGeneralResult.isFailed()) {
                throw new RuntimeException(serverGeneralResult.getMessage());
            }
            GeneralResult<CloudVendor> vendorResult = cmpCloudVendorService.detail(serverGeneralResult.getData().getVendorId());
            if (vendorResult.isFailed()) {
                throw new RuntimeException(vendorResult.getMessage());
            }
            Map<String, Object> item = new LinkedHashMap<>();
            if (!"".equals(monitorHostIpmi.getUsername())) {
                item.put("user", monitorHostIpmi.getUsername());
            }
            if (!"".equals(monitorHostIpmi.getPassword())) {
                item.put("pass", new AESEncryptor().decrypt(monitorHostIpmi.getPassword(), null));
            }
            item.put("driver", "LAN_2_0");
            item.put("collectors", Arrays.asList("chassis"));
            data.put(monitorHostIpmi.getHostId(), item);
            JSONObject ipmiData = new JSONObject();
            JSONObject module = new JSONObject();
            module.put("module", String.valueOf(monitorHostIpmi.getHostId()));
            module.put("vendor_id", String.valueOf(vendorResult.getData().getId()));
            module.put("hostname", serverGeneralResult.getData().getName());
            module.put("host_type", "openstack");
            ipmiData.put("targets", Arrays.asList(monitorHostIpmi.getIpmiIp()));
            ipmiData.put("labels", module);
            targets.add(ipmiData);
        }
        root.put("modules", data);
        // 不要在方法上抛出异常,使用catch捕捉
        FileWriter writer = null;
        try {
            Yaml yaml = new Yaml();
            writer = new FileWriter(this.storageConfig.getIpmiconfigFile());
            writer.write(yaml.dumpAsMap(root));
            writer.flush();
            writer.close();
            //重新生成targets.json文件
            //对json进行格式化
            String targetsData = JSON.toJSONString(targets, SerializerFeature.PrettyFormat, SerializerFeature.WriteMapNullValue,
                    SerializerFeature.WriteDateUseDateFormat);
            FileUtils.writeStringToFile(this.storageConfig.getIpmiTargetsFile(), targetsData, "UTF-8");
            reload();
        } catch (Exception e) {
            log.error("重新生成config.yml文件失败!", e);
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    log.error(e.getMessage(), e);
                }
            }
        }
    }

    @Override
    public GeneralResult unbind(Long hostId, Long userId) {
        MonitorHostIpmi hostIpmi = monitorHostIpmiRepository.query(hostId);
        if (null == hostIpmi) {
            return new GeneralResult<>(true, "当前宿主机未配置ipmi信息，无需解绑");
        }
        String path = MonitorHostIpmi.class.getSimpleName() + "_" + hostIpmi.getId();
        try (AutoCloseLock acLock = lockFactory.getACLock(path)) {
            Assert.isTrue(acLock.acquire(10, TimeUnit.SECONDS), "请求超时");
            hostIpmi.setMenderId(userId);
            hostIpmi.setDeleted(true);
            monitorHostIpmiRepository.update(hostIpmi);
            List<MonitorHostIpmi> monitorHostIpmis = monitorHostIpmiRepository.list();
            changeFile(monitorHostIpmis);
            reload();
            return new GeneralResult<>(true, "解绑成功");
        } catch (Exception e) {
            log.error("unbind the host for ipmi is error:", e);
            return new GeneralResult<>(false, "解绑ipmi信息失败");
        }

    }

    /**
     * 重新加载ipmiExporter服务
     */
    private void reload() {
        generateThreadPool.execute(() -> {
            try {
                Thread.sleep(5000L);
            } catch (InterruptedException e) {
                log.error(e.getMessage(), e);
                Thread.currentThread().interrupt();
            }
            String reloadUrl = ipmiExporterUrl + Constant.Monitor.RELOAD_URL;
            HttpClient client = new HttpClient();
            client.post(Maps.newHashMap(), Maps.newHashMap(), reloadUrl);
        });
    }

    @Override
    public GeneralResult detail(Long hostId) {
        try {
            MonitorHostIpmi monitorHostIpmi = monitorHostIpmiRepository.query(hostId);
            return new GeneralResult<>(true, monitorHostIpmi, "查询成功");
        } catch (Exception e) {
            log.error("Query ipmi information failed", e);
            return new GeneralResult<>(false, "查询失败");
        }
    }

}
