package com.bocloud.cms.service;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.common.config.StorageConfig;
import com.bocloud.cms.entity.*;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.enums.ResourceCategory;
import com.bocloud.cms.enums.ResourceType;
import com.bocloud.cms.intf.KvsService;
import com.bocloud.cms.intf.PrometheusRuleService;
import com.bocloud.cms.intf.PrometheusService;
import com.bocloud.cms.intf.ResourceAlarmStatusService;
import com.bocloud.cms.model.KvsModel;
import com.bocloud.cms.model.PrometheusAlarm;
import com.bocloud.cms.repository.ResourceAlarmStatusRepository;
import com.bocloud.cms.service.internal.mms.MmsKvsService;
import com.bocloud.pms.entity.Kvs;
import com.bocloud.pms.entity.KvsNode;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Pager;
import com.megatron.common.model.Param;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.MapTools;
import jakarta.servlet.http.HttpServletRequest;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import com.megatron.common.model.*;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Kvs key-value数据库监控
 * @author zhangyf
 * @version 1.0
 * @date 2022/2/21 18:01
 * @version 5.6.0
 */
@Service
@Slf4j
public class KvsServiceImpl implements KvsService, InitializingBean {


    @Autowired
    private MmsKvsService mmsKvsService;
    @Autowired
    private PrometheusService prometheusService;
    @Autowired
    private PrometheusRuleService prometheusRuleService;
    @Autowired
    private StorageConfig storageConfig;
    @Autowired
    private ResourceAlarmStatusRepository alarmStatusRepository;
    @Autowired
    private ResourceAlarmStatusService resourceAlarmStatusService;

    @Override
    public void init(Kvs kvs) throws Exception {
        switch (ResourceCategory.valueOf(kvs.getType())) {
            case Redis:
                generateRedisConf(kvs);
                break;
            case MongoDB:
                generateMongodbConf(kvs);
                break;
            case MemCache:
                generateMemcachedConf(kvs);
                break;
            default:
                log.warn("unsupported resource category " + kvs.getType());
                break;
        }
    }

    @Override
    public List<Kvs> listTargetObjects(List<String> targetIds, MonitorResourceType resourceType) {
        List<Integer> targetList = new ArrayList<>();
        Optional.ofNullable(targetIds).orElse(new ArrayList<>()).forEach(targetId -> targetList.add(Integer.valueOf(targetId)));
        Param param = new Param(MapTools.simpleMap("id", JSONObject.toJSONString(targetList).substring(1, JSONObject.toJSONString(targetList).length() - 1)), Sign.IN);

        List<Param> params = Collections.singletonList(param);
        GeneralResult result = mmsKvsService.list(new Pager(1, Integer.MAX_VALUE, params, null));
        Assert.isTrue(result.isSuccess(), result.getMessage());
        JSONObject data = (JSONObject) result.getData();
        return data.getJSONArray(Common.ROWS).toJavaList(Kvs.class);
    }

    @Override
    public Map<String, String> getPrometheusExprParam(Kvs targetObject, Metric metric) {
        Map<String, String> params = new HashMap<>(16, 0.75F);
        params.put("kvs_id", String.valueOf(targetObject.getId()));
        return params;
    }


    @Override
    public Kvs parseObjectFromAlert(PrometheusAlarm.PrometheusAlert alert, RuleGroup ruleGroup, Rule rule, Metric metric) {
        String kvsId = String.valueOf(alert.getLabels().get("kvs_id"));
        return mmsKvsService.detail(Long.parseLong(kvsId));
    }

    @Override
    public void destroy(Kvs kvs, String catalog) throws Exception {
        //删除prometheus targets文件
        File file = new File(new File(storageConfig.getPrometheusPath(), kvs.getType().toLowerCase()), "instance_" + kvs.getId() + ".json");
        if (file.exists() && !file.delete()) {
            log.error("file delete failed, path:{} " + file.getPath());
        }
        // 删除实例节点状态监听rule文件
        File ruleFile = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_kvs_" + kvs.getId() + ".yml");
        if (ruleFile.exists() && !ruleFile.delete()) {
            log.error("file delete failed, path:{} " + ruleFile.getPath());
        }
        prometheusService.reload();
        //TODO
    }

    @Override
    public void wrapAlarm(Alarm alarm, Kvs kvs) {
        alarm.setTarget(String.valueOf(kvs.getId()));
        if (kvs.getName()!=null & kvs.getManageIp()!=null){
            alarm.setTargetName(kvs.getName() + "(" + kvs.getManageIp() +")");
        }else{
            String targetName = kvs.getManageIp()==null ? kvs.getName() : kvs.getManageIp();
            alarm.setTargetName(targetName);
        }
        alarm.setOwnerId(kvs.getCreatorId());
    }

    @Override
    public boolean alarmEnable(RuleGroup ruleGroup, Kvs object) {
        return resourceAlarmStatusService.alarmEnable(ruleGroup.getResourceType(), String.valueOf(object.getId()), ruleGroup.getTenantId());
    }


    @Override
    public GeneralResult list(Pager pager, String catalog, RequestContext context, HttpServletRequest request) {

        GeneralResult result = mmsKvsService.list(pager, request.getHeader(Common.TOKEN));
        if (result.isFailed()) {
            return result;
        }
        JSONObject data = (JSONObject) result.getData();
        List<KvsModel> list = data.getJSONArray(Common.ROWS).toJavaList(KvsModel.class);
        if (list.isEmpty()) {
            data.put(Common.ROWS, list);
            return result;
        }
        List<String> targets = list.stream().map(m -> String.valueOf(m.getId())).collect(Collectors.toList());
        List<ResourceAlarmStatus> alarmStatuses = this.alarmStatusRepository.list(catalog, targets, context);
        for (KvsModel model : list) {
            //告警启用状态，默认开启
            boolean alarmEnable = !Optional.ofNullable(alarmStatuses).orElse(new ArrayList<>()).stream()
                    .filter(a -> a.getResourceId().equals(model.getId().toString()))
                    .findAny().isPresent();
            model.setAlarmEnable(alarmEnable);
        }
        data.put(Common.ROWS, list);
        return result;
    }

    /**
     * 生成Redis配置
     *
     * @param kvs
     * @throws IOException
     */
    private void generateRedisConf(Kvs kvs) throws IOException {
        // 创建prometheus targets文件
        List<JSONObject> list = kvs.getNodes().stream().filter(node -> !"sentinel".equalsIgnoreCase(node.getRole())).map(node -> {
            Map<String, String> labels = new HashMap<>(16, 0.75F);
            labels.put("kvs_id", String.valueOf(kvs.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("server", node.getAddress());
            labels.put("port", String.valueOf(node.getPort()));
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9273"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.Redis.name().toLowerCase()), "instance_" + kvs.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "redis node change event");
        JSONArray rules = new JSONArray();
        StringBuilder expr = new StringBuilder("changes((count(redis_clients{kvs_id='" + kvs.getId() + "'}) + 1 or absent(count(redis_clients{kvs_id='" + kvs.getId() + "'})))[3m:]) > 0");
        for (KvsNode node : kvs.getNodes()) {
            if ("sentinel".equalsIgnoreCase(node.getRole())) {
                continue;
            }
            expr.append(" or changes((count(redis_clients{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',replication_role='master'}) + 1 ");
            expr.append(" or absent(count(redis_clients{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',replication_role='master'})))[3m:]) > 0 ");
            expr.append(" or changes((count(redis_clients{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',replication_role='slave'}) + 1 ");
            expr.append(" or absent(count(redis_clients{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',replication_role='slave'})))[3m:]) > 0 ");
        }
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.Redis.name());
        labels.put("id", kvs.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "redis node change event");
        rule.put("expr", expr.toString());
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "redis instance(" + kvs.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_kvs_" + kvs.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    /**
     * 生成Mongodb配置
     *
     * @param kvs
     * @throws IOException
     */
    private void generateMongodbConf(Kvs kvs) throws IOException {
        // 创建prometheus targets文件
        List<JSONObject> list = kvs.getNodes().stream().filter(node -> !"arbiter".equalsIgnoreCase(node.getRole())).map(node -> {
            Map<String, String> labels = new HashMap<>(16, 0.75F);
            labels.put("kvs_id", String.valueOf(kvs.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("hostname", node.getAddress() + ":" + node.getPort());
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9273"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceCategory.MongoDB.name().toLowerCase()), "instance_" + kvs.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "mongodb node change event");
        JSONArray rules = new JSONArray();
        StringBuilder expr = new StringBuilder("changes((count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "'}) + 1 or absent(count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "'})))[3m:]) > 0 ");
        for (KvsNode node : kvs.getNodes()) {
            if ("arbiter".equalsIgnoreCase(node.getRole())) {
                continue;
            }
            expr.append(" or changes((count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',state='PRIMARY'}) + 1 ");
            expr.append(" or absent(count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',state='PRIMARY'})))[3m:]) > 0 ");
            expr.append(" or changes((count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',state='SECONDARY'}) + 1 ");
            expr.append(" or absent(count(mongodb_uptime_ns{kvs_id='" + kvs.getId() + "',node_id='" + node.getId() + "',state='SECONDARY'})))[3m:]) > 0");
        }
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.MongoDB.name());
        labels.put("id", kvs.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "mongodb node change event");
        rule.put("expr", expr.toString());
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "mongodb instance(" + kvs.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_kvs_" + kvs.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }

    /**
     * 生成Memcached配置
     *
     * @param kvs
     * @throws IOException
     */
    private void generateMemcachedConf(Kvs kvs) throws IOException {
        // 创建prometheus targets文件
        List<JSONObject> list = kvs.getNodes().stream().map(node -> {
            Map<String, Object> labels = new HashMap<>(16, 0.75F);
            labels.put("kvs_id", String.valueOf(kvs.getId()));
            labels.put("node_id", String.valueOf(node.getId()));
            labels.put("hostname", node.getAddress() + ":" + node.getPort());
            JSONObject config = new JSONObject();
            config.put("targets", Arrays.asList(node.getManageIp() + ":9273"));
            config.put("labels", labels);
            return config;
        }).collect(Collectors.toList());
        JSONArray array = new JSONArray();
        array.addAll(list);
        FileUtils.write(new File(new File(storageConfig.getPrometheusPath(), ResourceType.MEMCACHED.name().toLowerCase()), "instance_" + kvs.getId() + ".json"), array.toJSONString(), "UTF-8");

        // 创建prometheus rule 文件，用于监听服务节点状态是否变化
        JSONObject obj = new JSONObject();
        JSONArray groups = new JSONArray();
        JSONObject group = new JSONObject();
        group.put("name", "memcached node change event");
        JSONArray rules = new JSONArray();
        Map<String, Object> labels = Maps.newHashMap();
        labels.put("alertRuleType", "nodeEvent");
        labels.put("category", ResourceCategory.MemCache.name());
        labels.put("id", kvs.getId());
        JSONObject rule = new JSONObject();
        rule.put("alert", "memcached node change event");
        rule.put("expr", "changes((count(memcached_uptime{kvs_id='" + kvs.getId() + "'}) + 1 or absent(count(memcached_uptime{kvs_id='" + kvs.getId() + "'})))[3m:]) > 0 ");
        rule.put("for", "1s");
        rule.put("labels", labels);
        rule.put("annotations", MapTools.simpleMap("description", "memcached instance(" + kvs.getName() + ") node has changed"));
        rules.add(rule);
        group.put("rules", rules);
        groups.add(group);
        obj.put("groups", groups);
        File file = new File(storageConfig.getPrometheusRulesPath(), "alert_rule_event_kvs_" + kvs.getId() + ".yml");
        prometheusRuleService.generateFile(Lists.newArrayList(obj), Lists.newArrayList(file));
    }


    @Override
    public void afterPropertiesSet() throws Exception {
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_KVS_REDIS, this);
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_KVS_MONGODB, this);
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_KVS_MEMCACHED, this);
    }

}
