package com.bocloud.cms.service;

import com.bocloud.cms.entity.Metric;
import com.bocloud.cms.intf.MetricService;
import com.bocloud.cms.repository.MetricRepository;
import com.google.common.collect.Lists;
import com.megatron.common.model.*;
import com.megatron.common.utils.GridHelper;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;
import io.jsonwebtoken.lang.Assert;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * 监控指标
 *
 * @author fanzhenxi
 * @version 5.6.0
 * @date 2022/2/21 18:01
 */
@Service
@Slf4j
public class MetricServiceImpl implements MetricService, InitializingBean {

    private final LockFactory lockFactory;
    private final MetricRepository metricRepository;

    private final Map<Long, Metric> metricCache = new HashMap<>(16, 0.75F);

    public MetricServiceImpl(LockFactory lockFactory, MetricRepository metricRepository) {
        this.lockFactory = lockFactory;
        this.metricRepository = metricRepository;
    }

    @Override
    public GeneralResult list(String resourceType) {
        Map<String, Object> result = new LinkedHashMap<>();
        Map<String, List<Metric>> maps = new LinkedHashMap();
        List<String> title = new ArrayList<>();
        try {
            List<Metric> metrics = metricRepository.listMetrics(resourceType);
            for (Metric metric : metrics) {
                if (!maps.containsKey(metric.getTag())) {
                    title.add(metric.getTag());
                    List<Metric> list = Lists.newArrayList();
                    list.add(metric);
                    maps.put(metric.getTag(), list);
                } else {
                    List<Metric> list = maps.get(metric.getTag());
                    list.add(metric);
                    maps.put(metric.getTag(), list);
                }
            }
            result.put("title", title);
            result.put("content", maps);
        } catch (Exception e) {
            log.error("Metrics list error!", e);
            return new GeneralResult<>(false, "查询失败");
        }
        return new GeneralResult<>(true, result, "查询成功");
    }

    @Override
    public Metric query(Long id) {
        Metric metric = metricCache.get(id);
        if (null != metric) {
            return metric;
        }
        metric = metricRepository.get(Metric.class, id);
        if (null != metric) {
            metricCache.put(id, metric);
        }
        return metric;
    }

    @Override
    public GeneralResult monitorMetricList(Pager pager) {
        GridBean gridBean;
        try {
            // 判断监控资源类型是否存在 不存在则直接返回
            if (pager.getParams().size() == 0) {
                return new GeneralResult<>(false, "未找到该监控资源类型!");
            }
            int total = this.metricRepository.metricCount(pager.getParams());
            List<Metric> metricList = this.metricRepository.metricList(pager);
            gridBean = GridHelper.getBean(pager.getPage(), pager.getRows(), total, metricList);
            return new GeneralResult<>(true, gridBean, "查询成功!");
        } catch (Exception e) {
            log.error("find monitor metric list fail:", e);
            return new GeneralResult<>(false, "查询失败!");
        }
    }

    @Override
    public GeneralResult create(Metric metric, RequestContext context) {
        try {
            Assert.notNull(metric, "监控指标项不能为空！");
            Param params = new Param();
            params.setSign(Sign.EQ);
            // 同一类型下 监控指标名称不可重复
            Map<String, Object> param = MapTools.simpleMap("name", metric.getName());
            param.put("resource_type", metric.getResourceType());
            params.setParam(param);
            int nameCount = this.metricRepository.metricCount(Collections.singletonList(params));
            if (nameCount > 0) {
                return new GeneralResult<>(false, "新增失败,同一类型下的监控指标名称不可重复!");
            }
            // 同一类型下 监控指标表达式不可重复
            Map<String, Object> paramExpr = MapTools.simpleMap("expr", metric.getExpr());
            paramExpr.put("resource_type", metric.getResourceType());
            params.setParam(paramExpr);
            int exprCount = this.metricRepository.metricCount(Collections.singletonList(params));
            if (exprCount > 0) {
                return new GeneralResult<>(false, "新增失败,同一类型下的监控指标名称不可重复!");
            }
            // 默认新增的监控指标项不开启告警
            metric.setStatus(false);
            // 默认新增的监控指标项是启用状态
            metric.setActive(true);
            this.metricRepository.save(metric);
            this.initMetrics();
            return new GeneralResult<>(true, "新增监控指标项成功！");
        } catch (Exception e) {
            log.error("create monitor metric fail:", e);
            return new GeneralResult<>(false, "新增监控指标项失败！");
        }
    }

    @Override
    public GeneralResult modify(Metric metric, RequestContext context) {
        String path = Metric.class.getSimpleName() + "_" + metric.getId();
        try (AutoCloseLock acLock = lockFactory.getACLock(path)) {
            Assert.isTrue(acLock.acquire(10, TimeUnit.SECONDS), "请求超时");
            Param params = new Param();
            params.setSign(Sign.EQ);
            Metric dbMetric = this.metricRepository.get(Metric.class, metric.getId());
            Assert.notNull(dbMetric, " 数据不存在");
            // 同一类型下 监控指标名称不可重复
            if (!metric.getName().equals(dbMetric.getName())) {
                Map<String, Object> param = MapTools.simpleMap("name", metric.getName());
                param.put("resource_type", metric.getResourceType());
                params.setParam(param);
                int nameCount = this.metricRepository.metricCount(Collections.singletonList(params));
                if (nameCount > 0) {
                    return new GeneralResult<>(false, "修改失败,同一类型下的监控指标名称不可重复!");
                }
            }
            // 同一类型下 监控指标表达式不可重复
            if (!metric.getExpr().equals(dbMetric.getExpr())) {
                Map<String, Object> paramExpr = MapTools.simpleMap("expr", metric.getExpr());
                paramExpr.put("resource_type", metric.getResourceType());
                params.setParam(paramExpr);
                int exprCount = this.metricRepository.metricCount(Collections.singletonList(params));
                if (exprCount > 0) {
                    return new GeneralResult<>(false, "修改失败,同一类型下的监控指标名称不可重复!");
                }
            }
            // 默认监控指标项是启用状态
            metric.setActive(true);
            metric.setStatus(dbMetric.getStatus());
            metric.setResourceType(dbMetric.getResourceType());
            this.metricRepository.update(metric);
            this.initMetrics();
            return new GeneralResult<>(true, "修改成功!");
        } catch (Exception e) {
            log.error("modify monitor metric fail:", e);
            return new GeneralResult<>(false, "修改失败!");
        }
    }

    @Override
    public GeneralResult remove(Long id, RequestContext context) {
        String path = Metric.class.getSimpleName() + "_" + id;
        try (AutoCloseLock acLock = lockFactory.getACLock(path)) {
            Assert.isTrue(acLock.acquire(10, TimeUnit.SECONDS), "请求超时");
            Metric query = this.metricRepository.get(Metric.class, id);
            if (query == null) {
                return new GeneralResult<>(false, "删除失败,未查询到任何数据!");
            }
            // 将指标原始状态和启用状态全部修改为false -> 删除状态
            this.metricRepository.deletedMetric(id);
            this.initMetrics();
            return new GeneralResult<>(true, "删除成功!");
        } catch (Exception e) {
            log.error("deleted monitor metric fail:", e);
            return new GeneralResult<>(false, "删除失败!");
        }
    }

    @Override
    public GeneralResult batchRemove(List<Long> ids, RequestContext context) {
        int rightCount = 0;
        int wrongCount = 0;
        try {
            org.springframework.util.Assert.notEmpty(ids, " 至少选择一个告警规则删除");
            for (Long id : ids) {
                this.remove(id, context);
                rightCount++;
            }
        } catch (Exception e) {
            log.error("monitor metrics batch delete error!", e);
            return new GeneralResult<>(false, "批量删除告警规则失败:" + e.getMessage());
        }
        this.initMetrics();
        return new GeneralResult<>(true, "成功删除【" + rightCount + "】个，失败【" + wrongCount + "】个");
    }

    @Override
    public GeneralResult getByResourceType(String resourceType) {
        Assert.notNull(resourceType, "监控指标类型不能为空！");
        List<Metric> metrics = this.metricRepository.getByResourceType(resourceType);
        List<Long> metricIds = new ArrayList<>();
        metrics.forEach(metric -> {
            metricIds.add(metric.getId());
        });
        return new GeneralResult<>(true, metricIds, "查询成功!");
    }

    /**
     * 增删改方法执行后 将metric列表重新放入缓存
     */
    private void initMetrics() {
        metricCache.clear();
        List<Metric> list = metricRepository.list();
        for (Metric metric : list) {
            metricCache.put(metric.getId(), metric);
        }
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        List<Metric> list = metricRepository.list();
        for (Metric metric : list) {
            metricCache.put(metric.getId(), metric);
        }
    }
}
