package com.bocloud.cms.service;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cms.entity.Alarm;
import com.bocloud.cms.entity.Metric;
import com.bocloud.cms.entity.Rule;
import com.bocloud.cms.entity.RuleGroup;
import com.bocloud.cms.enums.MonitorResourceType;
import com.bocloud.cms.intf.MonitoringAble;
import com.bocloud.cms.intf.ResourceAlarmStatusService;
import com.bocloud.cms.model.PrometheusAlarm;
import com.bocloud.cms.model.ServerModel;
import com.bocloud.cms.service.internal.CmpCloudVendorService;
import com.bocloud.cms.service.internal.CmpServerService;
import com.bocloud.ims.entity.CloudVendor;
import com.bocloud.ims.entity.Server;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Pager;
import com.megatron.common.model.Param;
import com.megatron.common.model.Sign;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.MapTools;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;

import java.util.*;

/**
 * Vmware宿主机
 *
 * @author zhangyf
 * @since 2023/05/26
 */
@RequiredArgsConstructor
@Service
@Slf4j
public class OpenStackHostServiceImpl implements MonitoringAble<Server>, InitializingBean {

    private final CmpServerService cmpServerService;
    private final ResourceAlarmStatusService resourceAlarmStatusService;
    private final CmpCloudVendorService cmpCloudVendorService;

    @Override
    public void afterPropertiesSet() throws Exception {
        MonitoringAbleRegistry.register(MonitorResourceType.MONITOR_OPENSTACK_HOST, this);
    }

    @Override
    public List<Server> listTargetObjects(List<String> targetIds, MonitorResourceType resourceType) {
        List<Integer> targetList = new ArrayList<>();
        Optional.ofNullable(targetIds).orElse(new ArrayList<>()).forEach(targetId -> targetList.add(Integer.valueOf(targetId)));
        Param param = new Param(MapTools.simpleMap("id", JSONObject.toJSONString(targetList).substring(1, JSONObject.toJSONString(targetList).length() - 1)), Sign.IN);
        List<Param> params = Collections.singletonList(param);
        GeneralResult result = cmpServerService.list(new Pager(1, Integer.MAX_VALUE, params, null));
        Assert.isTrue(result.isSuccess(), result.getMessage());
        JSONObject data = (JSONObject) result.getData();
        return data.getJSONArray(Common.ROWS).toJavaList(Server.class);
    }

    @Override
    public Map<String, String> getPrometheusExprParam(Server targetObject, Metric metric) {
        Map<String, String> params = new HashMap<>(16, 0.75F);
        GeneralResult<CloudVendor> vendorResult = cmpCloudVendorService.detail(targetObject.getVendorId());
        if (vendorResult.isFailed()) {
            throw new RuntimeException(vendorResult.getMessage());
        }
        // op宿主机链接状态特殊化处理
        if ("bocloud_openstack_host_success".equals(metric.getExpr())) {
            params.put("vendor_id", String.valueOf(vendorResult.getData().getId()));
            params.put("host_ip", targetObject.getManageIp());
            params.put("hypervisor_hostname", targetObject.getName());
        } else if (metric.getExpr().contains("node_")) {
            params.put("host_id", String.valueOf(targetObject.getId()));
        } else {
            params.put("vendor_id", String.valueOf(vendorResult.getData().getId()));
            // update 采集opHost时放入主机ip和vendorId
            params.put("instance", targetObject.getManageIp());
        }
        return params;
    }

    @Override
    public Server parseObjectFromAlert(PrometheusAlarm.PrometheusAlert alert, RuleGroup ruleGroup, Rule rule, Metric metric) {
        if (metric.getExpr().contains("node_")) {
            String id = String.valueOf(alert.getLabels().get("host_id"));
            GeneralResult<Server> result = cmpServerService.detail(Long.valueOf(id));
            Assert.isTrue(result.isSuccess(), result.getMessage());
            return JSON.parseObject(JSON.toJSONString(result.getData()), ServerModel.class);
            // add 如果为事件监控，则通过manageIp查询Openstack主机对象
        } else if (metric.getIsEvent()) {
            // opHost为manageIp opVM为managerIp
            String manageIp = String.valueOf(alert.getLabels().get("host_ip"));
            List<Param> params = new ArrayList<>(1);
            params.add(new Param(MapTools.simpleMap("manageIp", manageIp), Sign.EQ));
            GeneralResult result = cmpServerService.list(new Pager(1, 1, params, null));
            Assert.isTrue(result.isSuccess(), result.getMessage());
            JSONObject data = (JSONObject) result.getData();
            List<Server> list = data.getJSONArray(Common.ROWS).toJavaList(Server.class);
            Assert.isTrue(list != null && !list.isEmpty(), "not found server,manageIp:" + manageIp);
            return list.get(0);
        } else {
            String vendorId = String.valueOf(alert.getLabels().get("vendor_id"));
            String name = String.valueOf(alert.getLabels().get("hypervisor_hostname"));
            List<Param> params = new ArrayList<>(2);
            params.add(new Param(MapTools.simpleMap("vendorId", vendorId), Sign.EQ));
            params.add(new Param(MapTools.simpleMap("name", name), Sign.EQ));
            GeneralResult result = cmpServerService.list(new Pager(1, 1, params, null));
            Assert.isTrue(result.isSuccess(), result.getMessage());
            JSONObject data = (JSONObject) result.getData();
            List<Server> list = data.getJSONArray(Common.ROWS).toJavaList(Server.class);
            Assert.isTrue(list != null && !list.isEmpty(), "not found server,name:" + name);
            return list.get(0);
        }
    }

    @Override
    public void destroy(Server object, String catalog) throws Exception {

    }

    @Override
    public void wrapAlarm(Alarm alarm, Server object) {
        if (object.getName() != null & object.getManageIp() != null) {
            alarm.setTargetName(object.getName() + "(" + object.getManageIp() + ")");
        } else {
            String targetName = object.getManageIp() == null ? object.getName() : object.getManageIp();
            alarm.setTargetName(targetName);
        }
        alarm.setVendorId(object.getVendorId());
        alarm.setOwnerId(object.getCreatorId());
        alarm.setTarget(String.valueOf(object.getId()));
    }

    @Override
    public boolean alarmEnable(RuleGroup ruleGroup, Server object) {
        return resourceAlarmStatusService.alarmEnable(ruleGroup.getResourceType(), String.valueOf(object.getId()), ruleGroup.getTenantId());
    }
}
